package net.knarcraft.clearonworldguard.config;

import com.sk89q.worldedit.bukkit.BukkitAdapter;
import com.sk89q.worldguard.WorldGuard;
import com.sk89q.worldguard.protection.managers.RegionManager;
import com.sk89q.worldguard.protection.regions.ProtectedRegion;
import com.sk89q.worldguard.protection.regions.RegionContainer;
import com.sk89q.worldguard.protection.regions.RegionQuery;
import net.knarcraft.clearonworldguard.ClearOnWorldGuard;
import org.bukkit.Bukkit;
import org.bukkit.World;
import org.bukkit.configuration.ConfigurationSection;
import org.bukkit.configuration.file.FileConfiguration;
import org.jetbrains.annotations.NotNull;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;
import java.util.logging.Level;

/**
 * A configuration that keeps track of loaded clear regions
 */
public class Configuration {

    private final RegionContainer regionContainer = WorldGuard.getInstance().getPlatform().getRegionContainer();
    private final RegionQuery regionQuery = regionContainer.createQuery();
    private final Map<World, Set<ProtectedRegion>> protectedRegions = new HashMap<>();
    private final boolean debugEnabled;

    /**
     * Instantiates a new configuration
     *
     * @param configuration <p>The file configuration to read</p>
     */
    public Configuration(@NotNull FileConfiguration configuration) {
        debugEnabled = Boolean.parseBoolean(configuration.getString("debug"));
        ConfigurationSection worldsSection = configuration.getConfigurationSection("clearRegions");
        if (worldsSection == null) {
            ClearOnWorldGuard.logger().log(Level.WARNING, "Unable to find clearRegions sections in the " +
                    "configuration file!");
            return;
        }
        Set<String> worlds = worldsSection.getKeys(false);

        for (String worldId : worlds) {
            // Parse every world identifier
            World world;
            try {
                UUID worldUUID = UUID.fromString(worldId);
                world = Bukkit.getWorld(worldUUID);
            } catch (IllegalArgumentException exception) {
                world = Bukkit.getWorld(worldId);
            }
            if (world == null) {
                ClearOnWorldGuard.logger().log(Level.WARNING, "The specified world: \"" + worldId + "\" is invalid.");
                continue;
            }

            loadRegions(world, worldsSection, worldId);
        }
    }

    /**
     * Loads clear regions for a world
     *
     * @param world         <p>The world to load regions for</p>
     * @param worldsSection <p>The configuration section listing all worlds and regions</p>
     * @param worldId       <p>The user-specified identifier for the currently processed world</p>
     */
    private void loadRegions(@NotNull World world, @NotNull ConfigurationSection worldsSection, @NotNull String worldId) {
        // Get a region manager for the world
        RegionManager regionManager = regionContainer.get(BukkitAdapter.adapt(world));
        if (regionManager == null) {
            ClearOnWorldGuard.logger().log(Level.WARNING, "Unable to get region manager for world: " +
                    world.getName());
            return;
        }

        Set<ProtectedRegion> regions = new HashSet<>();
        List<String> regionNames = worldsSection.getStringList(worldId);
        for (String regionName : regionNames) {
            // Register the region if it's valid
            ProtectedRegion region = regionManager.getRegion(regionName);
            if (region == null) {
                ClearOnWorldGuard.logger().log(Level.WARNING, "Region name " + regionName + " is invalid");
            } else {
                regions.add(region);
            }
        }
        protectedRegions.put(world, regions);
    }

    /**
     * Gets all protected regions for the given world
     *
     * @return <p>All protected regions for the given world</p>
     */
    public @NotNull Set<ProtectedRegion> getProtectedRegions(@NotNull World world) {
        return this.protectedRegions.get(world);
    }

    /**
     * Gets a region query
     *
     * @return <p>A region query</p>
     */
    public @NotNull RegionQuery getRegionQuery() {
        return regionQuery;
    }

    /**
     * Gets whether debug mode is enabled
     *
     * @return <p>True if debug mode is enabled</p>
     */
    public boolean debugEnabled() {
        return this.debugEnabled;
    }

}
