package net.knarcraft.clearonworldguard.listener;

import net.knarcraft.clearonworldguard.ClearOnWorldGuard;
import net.knarcraft.clearonworldguard.config.Configuration;
import org.bukkit.block.Block;
import org.bukkit.entity.Entity;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.block.BlockBreakEvent;
import org.bukkit.event.block.BlockBurnEvent;
import org.bukkit.event.block.BlockDispenseArmorEvent;
import org.bukkit.event.block.BlockDispenseEvent;
import org.bukkit.event.block.BlockDropItemEvent;
import org.bukkit.event.block.BlockExpEvent;
import org.bukkit.event.block.BlockFormEvent;
import org.bukkit.event.block.BlockIgniteEvent;
import org.bukkit.event.block.BlockPistonExtendEvent;
import org.bukkit.event.block.BlockPistonRetractEvent;
import org.bukkit.event.block.BlockPlaceEvent;
import org.bukkit.event.block.BlockSpreadEvent;
import org.bukkit.event.block.TNTPrimeEvent;
import org.jetbrains.annotations.NotNull;

import java.util.List;

/**
 * Listeners for block events
 */
public class BlockListener extends WorldGuardListener {

    /**
     * Instantiates a new WorldGuard listener
     *
     * @param configuration <p>The configuration to get regions and settings from</p>
     */
    public BlockListener(@NotNull Configuration configuration) {
        super(configuration);
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockItemDrop(@NotNull BlockDropItemEvent event) {
        // Block item drops for indirectly destroyed items
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from dropping as an item");
            event.getItems().clear();
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST)
    public void onBlockXP(@NotNull BlockExpEvent event) {
        // Prevent dropping XP in clear regions
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from dropping any XP");
            event.setExpToDrop(0);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onDispense(@NotNull BlockDispenseEvent event) {
        // Prevent dispensers from dispensing and droppers from dropping
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented dispenser " + event.getBlock() + " from dispensing");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onArmorDispense(@NotNull BlockDispenseArmorEvent event) {
        // Prevent dispensers from dispensing armors in clear regions
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented armor from dispensing from " + event.getBlock());
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBurn(@NotNull BlockBurnEvent event) {
        // Prevent blocks from burning in clear areas
        Block block = event.getBlock();
        if (isInClearRegion(block)) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from burning");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onIgnite(@NotNull BlockIgniteEvent event) {
        // Prevent ignition in clear areas, unless done by flint and steel
        Block block = event.getBlock();
        if (isInClearRegion(block) && event.getCause() != BlockIgniteEvent.IgniteCause.FLINT_AND_STEEL) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from igniting");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockSpread(@NotNull BlockSpreadEvent event) {
        // Prevent block spread in clear regions
        if (isInClearRegion(event.getSource())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from spreading");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockForm(@NotNull BlockFormEvent event) {
        // Prevent block forming in clear regions
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from forming");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onPistonExtend(@NotNull BlockPistonExtendEvent event) {
        // Prevent pistons from pushing blocks across the clear region border
        Block piston = event.getBlock();
        List<Block> blocks = event.getBlocks();
        for (Block block : blocks) {
            if (isInDifferentClearRegions(piston.getLocation(), block.getLocation())) {
                ClearOnWorldGuard.logDebugMessage("Prevented piston " + piston + " from extending");
                event.setCancelled(true);
                return;
            }
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onPistonRetract(@NotNull BlockPistonRetractEvent event) {
        // Prevent pistons from pulling blocks across the clear region border
        Block piston = event.getBlock();
        List<Block> blocks = event.getBlocks();
        for (Block block : blocks) {
            if (isInDifferentClearRegions(piston.getLocation(), block.getLocation())) {
                ClearOnWorldGuard.logDebugMessage("Prevented piston " + piston + " from retracting");
                event.setCancelled(true);
                return;
            }
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockPlace(@NotNull BlockPlaceEvent event) {
        // Prevent placing a block outside a clear region
        if (isInDifferentClearRegions(event.getPlayer().getLocation(), event.getBlockPlaced().getLocation())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + event.getBlock() + " from being placed");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBlockBreak(@NotNull BlockBreakEvent event) {
        // Prevent breaking a block outside a clear region, while in the clear region
        Block block = event.getBlock();
        if (isInDifferentClearRegions(event.getPlayer().getLocation(), block.getLocation())) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + block + " from being broken");
            event.setCancelled(true);
            return;
        }

        // Prevent item drops when the item is destroyed
        if (isInClearRegion(block)) {
            ClearOnWorldGuard.logDebugMessage("Prevented block " + block + " from dropping items");
            event.setDropItems(false);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onTNT(@NotNull TNTPrimeEvent event) {
        // Prevent TNT from being primed in a clear region, or by a player in a clear region
        Entity primingEntity = event.getPrimingEntity();
        Block block = event.getBlock();
        Block primingBlock = event.getPrimingBlock();
        if ((primingEntity != null && isInClearRegion(primingEntity)) || isInClearRegion(block) ||
                (primingBlock != null && isInClearRegion(primingBlock))) {
            ClearOnWorldGuard.logDebugMessage("Prevented TNT " + event.getBlock() + " from being primed by entity " +
                    primingEntity + " or block " + primingBlock);
            event.setCancelled(true);
        }
    }

}
