package net.knarcraft.clearonworldguard.listener;

import net.knarcraft.clearonworldguard.ClearOnWorldGuard;
import net.knarcraft.clearonworldguard.config.Configuration;
import net.knarcraft.clearonworldguard.manager.PlayerRegionTracker;
import net.knarcraft.clearonworldguard.property.Permission;
import org.bukkit.Location;
import org.bukkit.entity.Boat;
import org.bukkit.entity.Entity;
import org.bukkit.entity.EntityType;
import org.bukkit.entity.Player;
import org.bukkit.entity.Vehicle;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.entity.CreatureSpawnEvent;
import org.bukkit.event.entity.EntityChangeBlockEvent;
import org.bukkit.event.entity.EntityCombustEvent;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityDeathEvent;
import org.bukkit.event.entity.EntityExplodeEvent;
import org.bukkit.event.entity.EntityPickupItemEvent;
import org.bukkit.event.entity.ExplosionPrimeEvent;
import org.bukkit.event.entity.ProjectileLaunchEvent;
import org.bukkit.event.vehicle.VehicleDestroyEvent;
import org.bukkit.event.vehicle.VehicleMoveEvent;
import org.jetbrains.annotations.NotNull;
import org.spigotmc.event.entity.EntityMountEvent;

/**
 * Listeners for entity events
 */
public class EntityListener extends WorldGuardListener {

    /**
     * Instantiates a new WorldGuard listener
     *
     * @param configuration <p>The configuration to get regions and settings from</p>
     */
    public EntityListener(@NotNull Configuration configuration) {
        super(configuration);
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityExplode(@NotNull EntityExplodeEvent event) {
        // Prevent explosions in clear regions
        if (isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + event.getEntity() + " from exploding");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityMount(@NotNull EntityMountEvent event) {
        // Prevent entity mounting in clear regions
        Entity entity = event.getEntity();
        Entity mount = event.getMount();

        // Don't restrict NPCs
        if (entity instanceof Player player && player.hasMetadata("NPC")) {
            return;
        }

        if ((isInClearRegion(entity) || isInClearRegion(mount)) &&
                !entity.hasPermission(Permission.BYPASS_ENTITY_INTERACTION.toString())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + entity + " from mounting " + mount);
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityChangeBlock(@NotNull EntityChangeBlockEvent event) {
        // Prevent entities from changing blocks in clear regions
        if (isInClearRegion(event.getBlock())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + event.getEntity() + " from changing the block " +
                    event.getBlock());
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onCreatureSpawn(@NotNull CreatureSpawnEvent event) {
        // Prevent creatures from spawning, unless it's intentional
        CreatureSpawnEvent.SpawnReason spawnReason = event.getSpawnReason();
        if (isInClearRegion(event.getLocation()) && (event.getEntityType() != EntityType.ARMOR_STAND &&
                spawnReason != CreatureSpawnEvent.SpawnReason.COMMAND) &&
                spawnReason != CreatureSpawnEvent.SpawnReason.SPAWNER_EGG
        ) {
            if (spawnReason != CreatureSpawnEvent.SpawnReason.NATURAL &&
                    spawnReason != CreatureSpawnEvent.SpawnReason.BEEHIVE) {
                ClearOnWorldGuard.logDebugMessage("Prevented creature " + event.getEntity() + " from spawning with " +
                        "spawn reason " + event.getSpawnReason());
            }

            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityDamage(@NotNull EntityDamageByEntityEvent event) {
        // Prevent entity damage from players in clear regions
        if (!(event.getDamager() instanceof Player player)) {
            return;
        }

        if ((isInClearRegion(event.getEntity()) || PlayerRegionTracker.isInClearRegion(player)) &&
                !player.hasPermission(Permission.BYPASS_ENTITY_INTERACTION.toString())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + event.getEntity() + " from being damaged");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityDeath(@NotNull EntityDeathEvent event) {
        // Prevent mob loot and XP
        if (isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + event.getEntity() + " from dropping items");
            event.setDroppedExp(0);
            event.getDrops().clear();
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onItemPickup(@NotNull EntityPickupItemEvent event) {
        // Prevent item pickup in clear regions
        if (isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented entity " + event.getEntity() + " from picking up an item");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onVehicleDestroy(@NotNull VehicleDestroyEvent event) {
        // Prevent destroyed vehicles from dropping as items
        if (isInClearRegion(event.getVehicle())) {
            ClearOnWorldGuard.logDebugMessage("Prevented vehicle " + event.getVehicle() + " from dropping as an item");
            event.setCancelled(true);
            clearInventoriesAndRemoveRecursively(event.getVehicle());
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST)
    public void onVehicleMove(@NotNull VehicleMoveEvent event) {
        // Prevent chest minecarts or similar from leaving the clear region
        if (event.getFrom().equals(event.getTo())) {
            return;
        }

        Vehicle vehicle = event.getVehicle();

        // Player passengers are handled by the PlayerMoveEvent
        if (vehicle instanceof Boat) {
            for (Entity passenger : vehicle.getPassengers()) {
                if (passenger instanceof Player) {
                    return;
                }
            }
        }

        if (isInClearRegion(event.getFrom()) && isInDifferentClearRegions(event.getFrom(), event.getTo())) {
            ClearOnWorldGuard.logDebugMessage("Destroyed vehicle " + vehicle + " crossing a clear region border");
            clearInventoriesAndRemoveRecursively(vehicle);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onProjectileLaunch(@NotNull ProjectileLaunchEvent event) {
        // Prevent a player from using an XP bottle in a clear region
        if (isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented projectile " + event.getEntity());
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onExplosionPrime(@NotNull ExplosionPrimeEvent event) {
        // Prevent explosion from priming if any exploded blocks are in a clear region
        Location location = event.getEntity().getLocation();
        float radius = event.getRadius();
        if (isInClearRegion(location) ||
                isInClearRegion(location.clone().add(radius, radius, radius)) ||
                isInClearRegion(location.clone().add(-radius, radius, radius)) ||
                isInClearRegion(location.clone().add(-radius, radius, -radius)) ||
                isInClearRegion(location.clone().add(radius, radius, -radius)) ||
                isInClearRegion(location.clone().add(radius, -radius, radius)) ||
                isInClearRegion(location.clone().add(-radius, -radius, radius)) ||
                isInClearRegion(location.clone().add(-radius, -radius, -radius)) ||
                isInClearRegion(location.clone().add(radius, -radius, -radius))) {
            ClearOnWorldGuard.logDebugMessage("Prevented " + event.getEntity() + " from being primed");
            event.setCancelled(true);
            event.getEntity().remove();
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onCombust(@NotNull EntityCombustEvent event) {
        // Prevent entities in clear regions from being put on fire
        if (isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented " + event.getEntity() + " from combusting");
            event.setCancelled(true);
        }
    }

}
