package net.knarcraft.clearonworldguard.listener;

import com.sk89q.worldguard.protection.regions.ProtectedRegion;
import net.knarcraft.clearonworldguard.ClearOnWorldGuard;
import net.knarcraft.clearonworldguard.config.Configuration;
import net.knarcraft.clearonworldguard.event.EnterClearRegionEvent;
import net.knarcraft.clearonworldguard.event.ExitClearRegionEvent;
import net.knarcraft.clearonworldguard.manager.PlayerRegionTracker;
import net.knarcraft.clearonworldguard.property.Permission;
import org.bukkit.Bukkit;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.block.Block;
import org.bukkit.entity.HumanEntity;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.block.Action;
import org.bukkit.event.entity.PlayerDeathEvent;
import org.bukkit.event.entity.PlayerLeashEntityEvent;
import org.bukkit.event.player.PlayerBucketEmptyEvent;
import org.bukkit.event.player.PlayerChangedWorldEvent;
import org.bukkit.event.player.PlayerDropItemEvent;
import org.bukkit.event.player.PlayerInteractEntityEvent;
import org.bukkit.event.player.PlayerInteractEvent;
import org.bukkit.event.player.PlayerJoinEvent;
import org.bukkit.event.player.PlayerMoveEvent;
import org.bukkit.event.player.PlayerTeleportEvent;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.inventory.meta.SpawnEggMeta;
import org.jetbrains.annotations.NotNull;

import java.util.Set;

/**
 * Listeners for player events
 */
public class PlayerListener extends WorldGuardListener {

    /**
     * Instantiates a new WorldGuard listener
     *
     * @param configuration <p>The configuration to get regions and settings from</p>
     */
    public PlayerListener(@NotNull Configuration configuration) {
        super(configuration);
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onLeash(@NotNull PlayerLeashEntityEvent event) {
        // Prevent leashing of entities while in a clear region
        if (PlayerRegionTracker.isInClearRegion(event.getPlayer()) &&
                !event.getPlayer().hasPermission(Permission.BYPASS_ENTITY_INTERACTION.toString())) {
            ClearOnWorldGuard.logDebugMessage("Prevented " + event.getEntity() + " from being leashed by " +
                    event.getPlayer());
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onMove(@NotNull PlayerMoveEvent event) {
        Location from = event.getFrom();
        Location to = event.getTo();
        if (to == null || from.getWorld() == null) {
            return;
        }
        Player player = event.getPlayer();
        if (updateClearRegions(player, from, to)) {
            /* This must be done by overwriting the destination, instead of cancelling, to prevent the user from
               momentarily entering the game-mode of the clear region. */
            event.setTo(from);
        }
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onJoin(@NotNull PlayerJoinEvent event) {
        // Update clear regions of the player
        PlayerRegionTracker.replaceRegions(event.getPlayer(), getRegions(event.getPlayer().getLocation()));
    }

    @EventHandler(priority = EventPriority.MONITOR)
    public void onWorldChange(@NotNull PlayerChangedWorldEvent event) {
        // Update clear regions of the player
        PlayerRegionTracker.replaceRegions(event.getPlayer(), getRegions(event.getPlayer().getLocation()));
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onTeleport(@NotNull PlayerTeleportEvent event) {
        Location fromLocation = event.getFrom();
        Location toLocation = event.getTo();

        if (toLocation == null || fromLocation.equals(toLocation)) {
            return;
        }

        ClearOnWorldGuard.logDebugMessage("Detected teleporting player " + event.getPlayer());
        if (updateClearRegions(event.getPlayer(), fromLocation, toLocation)) {
            event.setCancelled(true);
            /* If this is missing, and the region puts players in creative mode, WorldGuard puts players in creative 
               mode until they move, even though the teleportation is cancelled. */
            Bukkit.getPluginManager().callEvent(new PlayerTeleportEvent(event.getPlayer(), toLocation, fromLocation));
        }
    }

    /**
     * Triggers any necessary events in order to change a player's clear regions
     *
     * @param player <p>The player that moved</p>
     * @param from   <p>The location the player moved from</p>
     * @param to     <p>The location the player moved to</p>
     * @return <p>True if whatever called this should be cancelled</p>
     */
    private boolean updateClearRegions(@NotNull Player player, @NotNull Location from, @NotNull Location to) {
        Set<ProtectedRegion> regionsEntered = getRegionsEntered(from, to);
        if (!regionsEntered.isEmpty()) {
            if (!player.getInventory().isEmpty() || player.isInsideVehicle() || !player.getPassengers().isEmpty()) {
                return true;
            }
            Bukkit.getPluginManager().callEvent(new EnterClearRegionEvent(player, regionsEntered));
        }

        Set<ProtectedRegion> regionsLeft = getRegionsExited(from, to);
        // Only trigger on the regions the player has entered
        regionsLeft.removeIf(region -> !PlayerRegionTracker.getClearRegions(player).contains(region));
        if (!regionsLeft.isEmpty()) {
            Bukkit.getPluginManager().callEvent(new ExitClearRegionEvent(player, regionsLeft));
        }
        return false;
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onItemDrop(@NotNull PlayerDropItemEvent event) {
        // Prevent a player from dropping an item while inside a clear region
        if (PlayerRegionTracker.isInClearRegion(event.getPlayer())) {
            ClearOnWorldGuard.logDebugMessage("Prevented player " + event.getPlayer() + " from dropping an item");
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST)
    public void onPlayerDeath(@NotNull PlayerDeathEvent event) {
        // Prevent deaths in clear regions from causing items to be kept or dropped
        if (PlayerRegionTracker.isInClearRegion(event.getEntity())) {
            ClearOnWorldGuard.logDebugMessage("Prevented player " + event.getEntity() + " from dropping or keeping " +
                    "items on death");
            event.setKeepInventory(false);
            event.setKeepLevel(true);
            event.getEntity().getInventory().clear();
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onInteract(@NotNull PlayerInteractEvent event) {
        // Prevent interacting with blocks outside the clear region, such as chests
        Player player = event.getPlayer();
        Block clicked = event.getClickedBlock();
        if (clicked != null && isInDifferentClearRegions(player.getLocation(), clicked.getLocation())) {
            ClearOnWorldGuard.logDebugMessage("Prevented player " + player + " from interacting across a clear" +
                    " region border");
            event.setCancelled(true);
            return;
        }

        // Block changing the creature in a spawner
        if (event.getAction() == Action.RIGHT_CLICK_BLOCK && clicked != null && clicked.getType() == Material.SPAWNER &&
                isInClearRegion(clicked) && !player.hasPermission(Permission.BYPASS_MOB_SPAWNING.toString())) {
            ClearOnWorldGuard.logDebugMessage("Prevented player " + player + " from interacting with mob spawner " +
                    clicked);
            event.setCancelled(true);
            return;
        }

        // Block usage of spawn eggs
        ItemStack item = event.getItem();
        if (item != null && PlayerRegionTracker.isInClearRegion(player)) {
            ItemMeta meta = item.getItemMeta();
            if (meta instanceof SpawnEggMeta && !player.hasPermission(Permission.BYPASS_MOB_SPAWNING.toString())) {
                ClearOnWorldGuard.logDebugMessage("Prevented player " + player + " from using the spawn egg " + item);
                event.setCancelled(true);
                return;
            }
        }
        
        /* Block usage of ender chests while in a clear region (the first check is this method takes care of the case
           where the chest is outside the clear region) */
        if (clicked != null && PlayerRegionTracker.isInClearRegion(player) &&
                event.getAction() != Action.LEFT_CLICK_BLOCK && clicked.getType() == Material.ENDER_CHEST) {
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onEntityInteract(@NotNull PlayerInteractEntityEvent event) {
        // Allow interactions with human NPCs
        if (event.getRightClicked() instanceof HumanEntity && !(event.getRightClicked() instanceof Player)) {
            return;
        }

        // Prevent entity interaction, such as trading or putting chests on donkeys
        if ((PlayerRegionTracker.isInClearRegion(event.getPlayer()) || isInClearRegion(event.getRightClicked())) &&
                !event.getPlayer().hasPermission(Permission.BYPASS_ENTITY_INTERACTION.toString())) {
            ClearOnWorldGuard.logDebugMessage("Prevented player " + event.getPlayer() +
                    " from interacting with entity " + event.getRightClicked());
            event.setCancelled(true);
        }
    }

    @EventHandler(priority = EventPriority.HIGHEST, ignoreCancelled = true)
    public void onBucketEmpty(@NotNull PlayerBucketEmptyEvent event) {
        if (PlayerRegionTracker.isInClearRegion(event.getPlayer())) {
            event.setCancelled(true);
        }
    }

}
