package net.knarcraft.clearonworldguard.listener;

import com.sk89q.worldedit.bukkit.BukkitAdapter;
import com.sk89q.worldguard.protection.regions.ProtectedRegion;
import net.knarcraft.clearonworldguard.ClearOnWorldGuard;
import net.knarcraft.clearonworldguard.config.Configuration;
import org.bukkit.Location;
import org.bukkit.World;
import org.bukkit.block.Block;
import org.bukkit.entity.Creature;
import org.bukkit.entity.Entity;
import org.bukkit.entity.Player;
import org.bukkit.event.Listener;
import org.bukkit.inventory.InventoryHolder;
import org.jetbrains.annotations.NotNull;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Queue;
import java.util.Set;
import java.util.logging.Level;

/**
 * A listener that listens to players entering and leaving WorldGuard regions
 */
public abstract class WorldGuardListener implements Listener {

    private final Configuration configuration;

    /**
     * Instantiates a new WorldGuard listener
     *
     * @param configuration <p>The configuration to get regions and settings from</p>
     */
    public WorldGuardListener(@NotNull Configuration configuration) {
        this.configuration = configuration;
    }

    /**
     * Recursively clears and removes entities connected together
     *
     * @param startEntity <p>The entity to start with</p>
     */
    protected void clearInventoriesAndRemoveRecursively(@NotNull Entity startEntity) {
        List<Entity> processed = new ArrayList<>();
        Queue<Entity> queue = new LinkedList<>();
        queue.add(startEntity);

        while (!queue.isEmpty()) {
            Entity entity = queue.remove();
            // Prevent duplicate processing
            if (processed.contains(entity)) {
                continue;
            }

            // Queue the entity's vehicle
            if (entity.isInsideVehicle()) {
                queue.add(entity.getVehicle());
            }
            // Queue the entity's passengers
            queue.addAll(entity.getPassengers());

            // Add any nearby creatures leashed to the entity
            Collection<Entity> nearbyEntities = entity.getWorld().getNearbyEntities(entity.getLocation(), 10, 10, 10);
            for (Entity nearby : nearbyEntities) {
                if (nearby instanceof Creature creature && creature.isLeashed() && creature.getLeashHolder().equals(entity)) {
                    queue.add(nearby);
                }
            }

            // Clear inventory if necessary
            if (entity instanceof InventoryHolder inventoryHolder) {
                inventoryHolder.getInventory().clear();
            }

            processed.add(entity);
        }

        // Eject and remove every entity
        for (Entity entity : processed) {
            if (!(entity instanceof Player)) {
                entity.eject();
                entity.remove();
            }
        }
    }

    /**
     * Checks whether the given locations are in different clear regions
     *
     * <p>If location1 is in a clear region and location2 is not, or if location1 is in a different clear region than
     * region2, or if region 1 is in an additional clear region, this will return true.</p>
     *
     * @param location1 <p>The first location to check</p>
     * @param location2 <p>The second location to check</p>
     * @return <p>True if the two locations are in different clear regions</p>
     */
    protected boolean isInDifferentClearRegions(@NotNull Location location1, @NotNull Location location2) {
        Set<ProtectedRegion> fromRegions = getOccupiedClearRegions(location1);
        Set<ProtectedRegion> toRegions = getOccupiedClearRegions(location2);

        // If the player is in one or more clear regions, clear unless the clear regions are the same
        return (!fromRegions.isEmpty() || !toRegions.isEmpty()) && !fromRegions.equals(toRegions);
    }

    /**
     * Gets the regions a player has newly entered
     *
     * @param location1 <p>The first location to check</p>
     * @param location2 <p>The second location to check</p>
     * @return <p>A list of all regions location2 is in, that location1 isn't in, or null if a world is null</p>
     */
    protected @NotNull Set<ProtectedRegion> getRegionsEntered(@NotNull Location location1, @NotNull Location location2) {
        Set<ProtectedRegion> fromRegions = getOccupiedClearRegions(location1);
        Set<ProtectedRegion> toRegions = getOccupiedClearRegions(location2);
        toRegions.removeAll(fromRegions);

        ClearOnWorldGuard.logDebugMessage("Calculated entered regions to: " + toRegions);

        return toRegions;
    }

    /**
     * Gets the regions a player has newly exited
     *
     * @param location1 <p>The first location to check</p>
     * @param location2 <p>The second location to check</p>
     * @return <p>A list of all regions location1 is in, that location2 isn't in, or null if a world is null</p>
     */
    protected @NotNull Set<ProtectedRegion> getRegionsExited(@NotNull Location location1, @NotNull Location location2) {
        Set<ProtectedRegion> fromRegions = getOccupiedClearRegions(location1);
        Set<ProtectedRegion> toRegions = getOccupiedClearRegions(location2);
        fromRegions.removeAll(toRegions);

        ClearOnWorldGuard.logDebugMessage("Calculated exited regions to: " + fromRegions);

        return fromRegions;
    }

    /**
     * Checks whether the given block is in a clear region
     *
     * @param entity <p>The entity to check</p>
     * @return <p>True if the block is in a clear region</p>
     */
    protected boolean isInClearRegion(@NotNull Entity entity) {
        return isInClearRegion(entity.getLocation());
    }

    /**
     * Checks whether the given block is in a clear region
     *
     * @param block <p>The block to check</p>
     * @return <p>True if the block is in a clear region</p>
     */
    protected boolean isInClearRegion(@NotNull Block block) {
        return isInClearRegion(block.getLocation());
    }

    /**
     * Checks whether the given location is in a clear region
     *
     * @param location <p>The location to check</p>
     * @return <p>True if the location is in a clear region</p>
     */
    protected boolean isInClearRegion(@NotNull Location location) {
        return !getOccupiedClearRegions(location).isEmpty();
    }

    /**
     * Gets all regions set as clear regions the given location currently occupies
     *
     * @param location <p>The location to get clear regions for</p>
     * @return <p>All clear regions the location occupies</p>
     */
    private @NotNull Set<ProtectedRegion> getOccupiedClearRegions(@NotNull Location location) {
        World playerWorld = location.getWorld();
        if (playerWorld == null) {
            ClearOnWorldGuard.logger().log(Level.WARNING, "Unable to check region change, as location " +
                    location + " has no world.");
            return new HashSet<>();
        }

        Set<ProtectedRegion> regions = new HashSet<>(getRegions(location));
        Set<ProtectedRegion> clearRegions = configuration.getProtectedRegions(playerWorld);
        regions.removeIf(region -> !clearRegions.contains(region));
        return regions;
    }

    /**
     * Gets all regions at the given location
     *
     * @param location <p>The location to get regions for</p>
     * @return <p>All regions at the location</p>
     */
    protected Set<ProtectedRegion> getRegions(@NotNull Location location) {
        return configuration.getRegionQuery().getApplicableRegions(BukkitAdapter.adapt(location)).getRegions();
    }

}
